#!/usr/bin/python3
#
# mlhub - Machine Learning Model Repository
#
# A command line tool for managing machine learning models.
#
# Support for a MLHub package developer.
#
# Copyright 2018-2021 (c) Graham.Williams@togaware.com All rights reserved.
#
# This file is part of mlhub.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
import os
import sys
import requests
import getpass
import subprocess
import re
import textwrap
import platform
from mlhub.utils import yes_or_no


# ----------------------------------------------------------------------
# Support Package Developers
# ----------------------------------------------------------------------

def load_key(path):
    """Load subscription key and endpoint from file."""
    key = None
    endpoint = None
    markchar = "'\" \t"
    with open(path, 'r') as file:
        for line in file:
            line = line.strip('\n')
            pair = line.split('=')
            if len(pair) == 2:
                k = pair[0].strip(markchar).lower()
                v = pair[1].strip(markchar)
                if k == 'key':
                    key = v
                elif k == 'endpoint':
                    endpoint = v
            elif not line.startswith('#'):
                line = line.strip(markchar)
                if line.startswith('http'):
                    endpoint = line
                else:
                    key = line
    return key, endpoint


def generalkey(key_file, priv_info, verbose=True, ask=True):
    """Load key, location, etc from file or ask user and save.

    The user is asked for an general key, location, etc.
     The provided information is saved into a file
    for future use. The contents of that file is the key.

    {"key":"abcd1234abcda4f2f6e9f565df34ef24","location":"eastaustralia"}

    """

    # Set up messages.

    msg_request = """\
Private information is required to access this service.
See the README for more details.
"""

    msg_found = f"""\
The following file has been found and is assumed to
contain the private information.

    {key_file}"""

    msg_saved = """
That information has been saved into the file:

    {}""".format(key_file)

    # Obtain the key/location/etc.

    if os.path.isfile(key_file) and os.path.getsize(key_file) > 0:
        if verbose:
            print(msg_found, file=sys.stderr)

        if ask:
            yes = yes_or_no("\nUse this private information ('d' to "
                            "display, 'n' to update)", third_choice=True)
        else:
            yes = True

        if yes == "d":
            with open(key_file, 'r') as handle:
                parsed = json.load(handle)
            print(f"\n{json.dumps(parsed, indent=2)}")
            yes = yes_or_no("\nUse this private information "
                            "(type 'n' to update)")

        if not yes:
            print("\n" + msg_request, file=sys.stderr)
            data = {}

            if any(isinstance(el, list) for el in priv_info):

                # private:
                #   Azure speech: key*, location
                #
                # In this case, priv_info = [[Azure speech, [key, location]]]

                for item in priv_info:
                    service = item[0]
                    nested_dic = {}

                    for elem in item[1]:
                        key_or_other = ask_info(elem, service)
                        nested_dic[elem] = key_or_other

                    data[service] = nested_dic

            # private:
            #   Azure speech:key*, location
            # In this case, the priv_info = [key, location]

            else:
                for item in priv_info:
                    key_or_other = ask_info(item, "")
                    data[item] = key_or_other

            # Write data into json file
            with open(key_file, "w") as outfile:
                json.dump(data, outfile)
            outfile.close()
            print(msg_saved, file=sys.stderr)
    else:
        print(msg_request, file=sys.stderr)

        if ask:
            data = {}

            if any(isinstance(el, list) for el in priv_info):

                # private:
                #   Azure speech:key*, location
                #
                # In this case, priv_info = [[Azure speech, [key, location]]]

                for item in priv_info:
                    service = item[0]
                    nested_dic = {}

                    for elem in item[1]:
                        key_or_other = ask_info(elem, service)
                        nested_dic[elem] = key_or_other

                    data[service] = nested_dic

            # private:
            #   Azure speech:key*, location
            # In this case, the priv_info = [key, location]

            else:
                for item in priv_info:
                    key_or_other = ask_info(item, "")
                    data[item] = key_or_other

            # Write data into json file
            with open(key_file, "w") as outfile:
                json.dump(data, outfile)
            outfile.close()
            print(msg_saved, file=sys.stderr)


def ask_info(item, service):
    if "*" in item:
        item = item.replace("*", "")
        key_or_other = ask_password(f"Please paste your {service} {item}: ")
    else:
        sys.stderr.write(f"Please paste your {item}: ")
        key_or_other = input()
    return key_or_other


def ask_password(prompt=None):
    if prompt is None:
        prompt = "Password: "
    return getpass.getpass(prompt=prompt)


# Send a request.


def azrequest(endpoint, url, subscription_key, request_data):
    """Send anomaly detection request to the Anomaly Detector API.

    If the request is successful, the JSON response is returned.

    Aim to generalise this to go into MLHUB to send request.
    """

    headers = {'Content-Type': 'application/json',
               'Ocp-Apim-Subscription-Key': subscription_key}

    response = requests.post(os.path.join(endpoint, url),
                             data=json.dumps(request_data),
                             headers=headers)

    if response.status_code == 200:
        return json.loads(response.content.decode("utf-8"))
    else:
        print(response.status_code)
        raise Exception(response.text)


def get_char():
    # From https://stackoverflow.com/questions/510357/
    # how-to-read-a-single-character-from-the-user
    # try to bypass python input()
    if is_windows():
        # For Windows-based systems
        import msvcrt  # If successful, we are on Windows
        return msvcrt.getch()
    else:
        # For POSIX-based systems (with termios & tty support)
        import tty
        import sys
        import termios  # Raises ImportError if unsupported.

        fd = sys.stdin.fileno()
        oldSettings = termios.tcgetattr(fd)

        try:
            tty.setcbreak(fd)
            answer = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, oldSettings)

        return answer


def mlask(prompt="Press Enter to continue", begin="", end=""):
    begin = "\n" if isinstance(begin, bool) and begin else begin
    end = "\n" if isinstance(end, bool) and end else end

    if isinstance(prompt, bool):
        sys.stderr.write("\n**Warning**\n"
                         "mlhub.pkg.mlask() parameter order updated.\n"
                         "Ignoring begin setting for now.\n"
                         "The source mlhub package needs to be updated."
                         "\n")
        prompt = "Press Enter to continue"

    sys.stdout.write(begin + prompt + ": ")
    sys.stdout.flush()
    get_char()
    sys.stdout.write(end + "\n")


def mlcat(title="", text="", delim="=", begin="", end="\n"):
    sep = delim * len(title) + "\n" if len(title) > 0 else ""
    ttl_sep = "\n" if len(title) > 0 else ""
    # Retain any extra line in the original text since fill() will
    # remove it.
    if len(text) and text[-1] == "\n":
        end = "\n" + end
    # Split into paragraphs, fill each paragraph, convert back to a
    # list of strings, and join them together as the text to be
    # printed.
    text = [l.strip() for l in list(map(textwrap.fill, text.split("\n\n")))]
    text = "\n\n".join(text)
    print(begin + sep + title + ttl_sep + sep + ttl_sep + text, end=end)


def is_linux():
    return platform.system() == "Linux"


def is_windows():
    return platform.system() == "Windows"


def is_mac():
    return platform.system() == "Darwin"


if is_linux():
    CTRLW = "using Ctrl-W"
elif is_mac():
    CTRLW = "using Cmd-W"
else:
    CTRLW = "when ready"


def mlpreview(fname,
              begin="\n",
              msg=f"Close the graphic window {CTRLW}.\n",
              previewer=None):
    if begin != "" or msg != "":
        print(begin + msg)
    if is_linux():
        if previewer is None:
            # Linux desktop default application for the file.
            previewer = "xdg-open"
        subprocess.Popen([previewer, fname])
    elif is_windows():
        # Windows automatically uses default program based on file extension.
        subprocess.Popen(["start", fname], shell=True)
    else:
        if previewer is None:
            previewer = "open"  # Perhaps a Mac?
        subprocess.Popen([previewer, fname])


# From Simon Zhao's azface package on github.

def is_url(url):
    """Check if url is a valid URL."""

    urlregex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+'
        r'(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    if re.match(urlregex, url) is not None:
        return True
    else:
        return False


def get_cmd_cwd():
    """Return the dir where model pkg command is invoked.

    For example, if `cd /temp; ml demo xxx`, then get_cmd_cwd()
    returns `/temp`.  It is used by model pkg developer, and is
    different from where the model pkg script is located.

    `CMD_CWD` is a environment variable passed by
    mlhub.utils.dispatch() when invoke model pkg script.
    """

    cmd_cwd = os.environ.get("_MLHUB_CMD_CWD", "")
    if cmd_cwd == "":
        cmd_cwd = os.getcwd()
    return cmd_cwd


def get_private(file_path="private.json", server=None):
    """Return a list of private information

    For example when call get_private()
    returns ["asdfghjkl", "westus"]. The first element is key,and the second
    is location. The order in the list is the same as the order in private
    row in MLHUB.yaml.
    If server is None, this function will return the first private information,
    otherwise it will return the specific one.

    """
    path = os.path.join(os.getcwd(), file_path)
    if os.path.exists(path):
        with open(path) as f:
            private_info = json.load(f)
            values = list(private_info.values())

            # The MLHub yaml includes
            #
            # private:
            #   Azure Speech: key*, location

            if any(isinstance(el, dict) for el in values):
                for item in values:
                    for i in list(item.values()):
                        if i == "":
                            sys.exit("Your private information is blank. "
                                     "Please run ml configure <model> to "
                                     "paste your private information.")
                if server is None:
                    return list(values[0].values())
                else:
                    if server in list(private_info.keys()):
                        return list(private_info[server].values())
                    else:
                        sys.exit("The server's name doesn't exist.\n"
                                 "Please make sure you have the correct name.")

            # private:key*, location
            # In this case, the values = [asdfghj(key), australia(location)]

            else:
                for item in values:
                    if item == "":
                        sys.exit("Your private information is blank. "
                                 "Please run ml configure <model> to "
                                 "paste your private information.")
                return values

    else:
        sys.exit("Please run ml configure <model> to "
                 "paste your private information.")



#!/usr/bin/python3
#
# mlhub - Machine Learning Model Repository
#
# A command line tool for managing machine learning models.
#
# Copyright 2018 (c) Graham.Williams@togaware.com All rights reserved.
#
# This file is part of mlhub.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the ""Software""), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import argparse
import logging
import mlhub.commands as commands
import mlhub.constants as constants
import mlhub.utils as utils
import os
import sys

# Expose exported functions.

from .pkg import get_cmd_cwd
from .pkg import get_private
from .pkg import mlask
from .pkg import mlpreview

# ----------------------------------------------------------------------
# Set up log.
# ----------------------------------------------------------------------

# Initialize log dir.  Ensure it exists.

utils.create_log_dir()

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Add file log handler to log all into a file

utils.add_log_handler(
    logger,
    logging.FileHandler(constants.LOG_FILE),
    constants.LOG_FILE_LEVEL,
    constants.LOG_FILE_FORMAT)

logger.info('---------- {} {} ----------'.format(os.path.basename(sys.argv[0]),
                                                 ' '.join(sys.argv[1:])))

# ----------------------------------------------------------------------
# Set up command line parser and dispatch commands.
# ----------------------------------------------------------------------


def main():
    """Main program for the command line script."""

    # --------------------------------------------------
    # Global option parser.  See mlhub.constants.OPTIONS
    # --------------------------------------------------

    logger.info("Create global option parser.")
    global_option_parser = argparse.ArgumentParser(
        add_help=False  # Disable -h or --help.  Use custom help msg instead.
    )
    utils.OptionAdder(global_option_parser, constants.OPTIONS).add_alloptions()

    # --------------------------------------------------
    # Parse version
    # --------------------------------------------------

    logger.info("Parse global options.")
    args, extras = global_option_parser.parse_known_args(sys.argv[1:])

    if args.debug:  # Add console log handler to log debug message to console
        logger.info('Enable printing out debug log on console.')
        utils.add_log_handler(
            logger,
            logging.StreamHandler(),
            logging.DEBUG,
            constants.LOG_CONSOLE_FORMAT)

    logger.debug(f'args: {args}, extra_args: {extras}')

    # Get the first positional argument.

    pos_args = [(i, arg)
                for i, arg in enumerate(sys.argv[1:])
                if not arg.startswith('-')]
    first_pos_arg_index, first_pos_arg = pos_args[0] if len(pos_args) != 0 else (None, None)
    logger.debug(f'First positional argument: {first_pos_arg}')

    if args.version:
        logger.info('Query version.')

        # --------------------------------------------------
        # Query the version of the model, for example
        #   $ ml rain -v
        #   $ ml -v rain
        # Otherwise, output the version of ml
        #   $ ml -v
        # --------------------------------------------------

        if first_pos_arg is not None:  # Print model version
            print(first_pos_arg, "version", utils.get_version(first_pos_arg))
        else:  # Print mlhub version
            print(constants.APP, "version", utils.get_version())

        return 0

    # --------------------------------------------------
    # Parse command line args for basic commands or model specific commands
    # --------------------------------------------------

    # Correct misspelled command if possible.

    if first_pos_arg is not None:

        # Only match basic commands since model pkg commands are more
        # specific which would be better to be checked after the model
        # pkg name is known.

        # 20200719 Check for aliases here as a temporary hack since
        # aliases are not currently dealt with in the call to
        # get_misspelled_command().

        if first_pos_arg == "remove":
            matched_cmd = "uninstall"
        else:
            matched_cmd = utils.get_misspelled_command(first_pos_arg,
                                                       list(constants.COMMANDS))

        if matched_cmd is not None:
            sys.argv[first_pos_arg_index + 1] = matched_cmd
            first_pos_arg = matched_cmd

    # Dispatch commands.

    if first_pos_arg is not None and first_pos_arg not in constants.COMMANDS:

        # Model specific commands, such as demo, display.

        logger.info("Parse model specific commands.")
        model_cmd_parser = argparse.ArgumentParser(
            prog=constants.CMD,
            parents=[global_option_parser],
            add_help=False  # Use custom help message
        )
        model_cmd_parser.add_argument('cmd', metavar='command')
        model_cmd_parser.add_argument('model')
        args, extras = model_cmd_parser.parse_known_args()
        logger.debug(f"args: {args}")
        logger.debug(f"extra_args: {extras}")

        # Simple help message for the model-specific command

        if '--help' in extras or '-h' in extras:
            logger.debug(f"Help for command '{args.cmd}' of '{args.model}'")
            utils.print_model_cmd_help(utils.load_description(args.model), args.cmd)
            print()
            return 0

        setattr(args, 'func', commands.dispatch)
        setattr(args, 'param', extras)

    else:

        # Basic commands, like install, readme.  See
        # mlhub.constants.COMMANDS

        logger.info("Parse basic common commands.")
        basic_cmd_parser = argparse.ArgumentParser(
            prog=constants.CMD,
            description="Access models from the ML Hub.",
            parents=[global_option_parser])
        subparsers = basic_cmd_parser.add_subparsers(title='subcommands', dest="cmd")
        utils.SubCmdAdder(subparsers, commands, constants.COMMANDS).add_allsubcmds()
        args = basic_cmd_parser.parse_args()
        logger.debug(f"args: {args}")

    # Print usage for incorrect argument

    if "func" not in args:
        utils.print_usage()
        return 0

    # Ensure we have a trailing slash on the mlhub.

    if args.mlhub is not None:
        constants.MLHUB = os.path.join(args.mlhub, "")

    if args.mlmetavar is not None:
        constants.CMD = args.mlmetavar

    # --------------------------------------------------
    # Dispatch commands
    # --------------------------------------------------

    try:

        args.func(args)

    except utils.MLInitCreateException as e:
        msg = "The below '{}' init folder cannot be created:\n  {}"
        utils.print_error_exit(msg, constants.APP, e.args[0])

    except utils.MLTmpDirCreateException as e:
        msg = "The below '{}' tmp folder cannot be created:\n  {}"
        utils.print_error_exit(msg, constants.APP, e.args[0])

    except utils.MalformedMLMFileNameException as e:
        msg = "Malformed {} file:\n  {}"
        utils.print_error_exit(msg, constants.EXT_MLM, e.args[0])

    except utils.MalformedYAMLException as e:
        name = e.args[0]
        if os.path.sep in name or '/' in name:
            msg = "Malformed YAML file:\n  {}"
        else:
            msg = "Malformed description for model package '{}'!"
        utils.print_error_exit(msg, e.args[0])

    except utils.ModelURLAccessException as e:
        msg = "URL access failed:\n  {}"
        utils.print_error_exit(msg, e.args[0])

    except utils.YAMLFileAccessException as e:
        msg = "YAML file access failed:\n  {}"
        utils.print_error_exit(msg, e.args[0])

    except utils.RepoAccessException as e:
        utils.print_error("Cannot access the ML Hub repository:\n  {}", e.args[0])
        if not args.quiet:  # Suggest check if any models installed, since mlhub repo not accessible
            utils.print_commands_suggestions_on_stderr('installed')
        sys.exit(1)

    except utils.ModelNotFoundOnRepoException as e:
        msg = "No model named '{}' was found on\n  {}"
        utils.print_error(msg, e.args[0], e.args[1])
        if not args.quiet:  # Suggest check if any models available, since specified model not found
            utils.print_commands_suggestions_on_stderr('available')
        sys.exit(1)

    except utils.ModelDownloadHaltException as e:
        msg = f"URL - '{e.args[0]}' failed:\n  {e.args[1]}"
        utils.print_error_exit(msg)

    except utils.DescriptionYAMLNotFoundException:
        msg = f"No '{utils.MLHUB_YAML}' description file "
        msg += f"found in '{utils.MLINIT}'."
        utils.print_error(msg)
        sys.exit(1)

    except utils.ModelNotInstalledException as e:
        msg = "model '{}' is not installed ({})."
        utils.print_error(msg, e.args[0], utils.get_init_dir())
        if not args.quiet:
            utils.print_commands_suggestions_on_stderr('installed',
                                                       'available',
                                                       'install')
        sys.exit(1)

    except utils.ModelInstalledException as e:
        msg = f"a model named '{e}' already exists."
        utils.print_error(msg)
        if not args.quiet:
            utils.print_commands_suggestions_on_stderr('installed',
                                                       'available',
                                                       'uninstall')
        sys.exit(1)

    except utils.ModelReadmeNotFoundException as e:
        msg = "The '{}' model does not have a '{}' file.\n"
        utils.print_error(msg, e.args[0], constants.README, e.args[1])
        sys.exit(1)

    except utils.UnsupportedScriptExtensionException as e:
        msg = "Could not determine an interpreter for extension '{}'"
        utils.print_error_exit(msg, e.args[0])

    except utils.CommandNotFoundException as e:
        msg = "The command '{}' was not found for this model '{}'."
        utils.print_error(msg, e.args[0], e.args[1])
        if not args.quiet:  # Suggest check if any available commands
            utils.print_commands_suggestions_on_stderr('commands')
        sys.exit(1)

    except utils.LackDependencyException as e:
        msg = "Required {} dependencies are not installed for this model: \n  ====> \033[31m{}\033[0m"
        utils.print_error(msg, 'R' if e.args[1] else 'Python',  e.args[0])
        if not args.quiet:  # Suggest install dependencies
            utils.print_commands_suggestions_on_stderr('configure')
        sys.exit(1)

    except utils.LackPrerequisiteException as e:
        msg = "Required pre-requisite not found: \n  ====> \033[31m{}\033[0m"
        utils.print_error(msg, e.args[0])
        if not args.quiet:  # Suggest install dependencies
            msg = "\nTo install required pre-requisites:\n\n  $ ml configure\n"
            utils.print_on_stderr(msg)
        sys.exit(1)

    except utils.DataResourceNotFoundException:
        msg = "Some data or model files required by the model package are missing!"
        utils.print_error(msg)
        if not args.quiet:  # Suggest download data
            utils.print_commands_suggestions_on_stderr('configure')
        sys.exit(1)

    except utils.MalformedPackagesDotYAMLException as e:
        msg = (f"No '{e.args[0]}' available for the package '{e.args[1]}'.")
        utils.print_error(msg)
        # Suggest AVAILABLE since specified model is not available.
        if not args.quiet:
            utils.print_commands_suggestions_on_stderr('available')
        sys.exit(1)

    except utils.ModelPkgInstallationFileNotFoundException as e:
        msg = "No such package file: {}\n  The model package may be broken!"
        utils.print_error_exit(msg, e.args[0])

    except utils.ModelPkgDependencyFileNotFoundException as e:
        msg = "Failed to get file dependency: {}\n"
        utils.print_error_exit(msg, e.args[0])

    except utils.ModelPkgDependencyFileTypeUnknownException as e:
        msg = "Unknown file dependency type: {}\n"
        utils.print_error_exit(msg, e.args[0])

    except utils.ConfigureFailedException as e:  # configure failed, then just quit
        msg = "An error was encountered:\n{}\n"
        utils.print_error_exit(msg, e.args[0])

    except utils.InstallFailedException as e:
        msg = "An error was encountered:\n{}\n"
        utils.print_error_exit(msg, e.args[0])

    except (KeyboardInterrupt, EOFError):  # Catch Ctrl-C and Ctrl-D
        print()
        sys.exit(1)


if __name__ == "__main__":
    main()
